package TextualTemplates.TextualTemplates;

import DataStructures.Array.Structures.Array;
import DataStructures.Array.Structures.Data;
import DataStructures.Array.Structures.DataReference;
import DataStructures.Array.Structures.Structure;
import TextualTemplates.LinkedListNodeFunctions.LinkedListNodeNodes;
import lists.DynamicArrayCharacters.Structures.*;
import lists.LinkedListCharacters.Structures.*;
import lists.LinkedListStrings.Structures.LinkedListStrings;
import references.references.*;

import static DataStructures.Array.Arrays.Arrays.*;
import static DataStructures.Array.Structures.Structures.*;
import static JSON.Parser.Parser.ReadJSON;
import static JSON.Writer.Writer.*;
import static TextualTemplates.LinkedListNodeFunctions.LinkedListNodeFunctions.*;
import static arrays.arrays.arrays.*;
import static lists.DynamicArrayCharacters.DynamicArrayCharactersFunctions.DynamicArrayCharactersFunctions.*;
import static lists.LinkedListCharacters.LinkedListCharactersFunctions.LinkedListCharactersFunctions.*;
import static lists.LinkedListStrings.LinkedListStringsFunctions.LinkedListStringsFunctions.*;
import static references.references.references.*;
import static strings.strings.strings.*;

public class TextualTemplates {

    public static boolean IsValidTemplate(char [] template){
        StringReference message;
        boolean success;
        LinkedListStrings tokens;
        StringReference[] ts;
        Node root;

        tokens = CreateLinkedListString();
        message = new StringReference();

        success = GenerateTokensFromTemplate(template, tokens, message);

        if(success) {
            root = new Node();

            ts = LinkedListStringsToArray(tokens);

            success = ParseTemplate(ts, root, message);
        }

        if(!success){
            FreeStringReference(message);
        }

        return success;
    }

    public static boolean GenerateTokensFromTemplate(char [] template, LinkedListStrings tokens, StringReference message){
        boolean success, found;
        double i, j, l;
        char pc, c;
        char [] cs, a;
        DynamicArrayCharacters da;

        success = true;
        cs = new char [4];

        da = CreateDynamicArrayCharacters();

        pc = 'x';
        for(i = 0d; i < template.length && success;){
            c = template[(int)i];

            if(c != '{') {
                DynamicArrayAddCharacter(da, c);
                i = i + 1d;
            }else{
                FillString(cs, 'x');
                for(j = 0; i+j < template.length && j < cs.length; j = j + 1d){
                    cs[(int)j] = template[(int)(i+j)];
                }

                if(
                    StringsEqual(cs, "{use".toCharArray()) ||
                    StringsEqual(cs, "{end".toCharArray()) ||
                    StringsEqual(cs, "{pri".toCharArray()) ||
                    StringsEqual(cs, "{for".toCharArray()) ||
                    StringsEqual(cs, "{if ".toCharArray()) ||
                    StringsEqual(cs, "{els".toCharArray())
                ){
                    if(pc != '\\'){
                        // Find end.
                        found = false;
                        for(l = 0; i+l < template.length && !found; l = l + 1d){
                            if(template[(int)(i + l)] == '}'){
                                found = true;
                            }
                        }

                        if(found){
                            if(da.length > 0d){
                                a = DynamicArrayCharactersToArray(da);
                                LinkedListAddString(tokens, a);
                                FreeDynamicArrayCharacters(da);
                                da = CreateDynamicArrayCharacters();
                            }

                            for(j = 0d; j < l; j = j + 1d){
                                DynamicArrayAddCharacter(da, template[(int)(i + j)]);
                            }

                            a = DynamicArrayCharactersToArray(da);
                            LinkedListAddString(tokens, a);
                            FreeDynamicArrayCharacters(da);
                            da = CreateDynamicArrayCharacters();

                            i = i + l;
                        }else{
                            success = false;
                            message.string = "Template command found, but not ended properly.".toCharArray();
                        }
                    }else{
                        DynamicArrayAddCharacter(da, c);
                        i = i + 1d;
                    }
                }else{
                    DynamicArrayAddCharacter(da, c);
                    i = i + 1d;
                }
            }

            pc = c;
        }

        if(da.length > 0d){
            a = DynamicArrayCharactersToArray(da);
            LinkedListAddString(tokens, a);
        }

        FreeDynamicArrayCharacters(da);

        return success;
    }

    public static boolean GenerateDocument(char [] template, char [] json, StringReference document, StringReference message){
        DataReference data;
        boolean success;

        data = new DataReference();

        success = ReadJSON(json, data, message);

        if(success) {
            success = GenerateDocumentBasedOnDataStructure(template, data.data, document, message);
        }

        return success;
    }

    public static boolean GenerateDocumentBasedOnDataStructure(char [] template, Data data, StringReference document, StringReference message){
        LinkedListCharacters ll;
        boolean success;
        LinkedListStrings tokens;
        StringReference[] ts;
        Node root;
        Array parents;

        tokens = CreateLinkedListString();
        parents = CreateArray();

        success = GenerateTokensFromTemplate(template, tokens, message);

        if(success) {
            root = new Node();

            ts = LinkedListStringsToArray(tokens);
            FreeLinkedListString(tokens);

            success = ParseTemplate(ts, root, message);

            if(success) {
                ll = CreateLinkedListCharacter();

                success = GenerateDocumentFromBlock(root, data, parents, ll, message);

                if(success) {
                    document.string = LinkedListCharactersToArray(ll);
                    FreeLinkedListCharacter(ll);
                }
            }
        }

        return success;
    }

    public static boolean GenerateDocumentFromBlock(Node root, Data data, Array parents, LinkedListCharacters ll, StringReference message) {
        LinkedListNodeNodes n;
        boolean success;

        n = root.nodes.first;
        success = true;

        for(; !n.end && success; ){
            success = GenerateDocumentFromNode(n.value, data, parents, ll, message);
            n = n.next;
        }

        return success;
    }

    public static boolean GenerateDocumentFromNode(Node n, Data data, Array parents, LinkedListCharacters ll, StringReference message) {
        boolean success;
        BooleanReference found;
        char [] valueString;
        Data value;
        double i, p;
        Structure parentStruct;

        success = true;
        found = new BooleanReference();

        if(StringsEqual(n.type, "block".toCharArray())){
            success = GenerateDocumentFromBlock(n, data, parents, ll, message);
        }else if(StringsEqual(n.type, "use".toCharArray())){
            success = GenerateDocumentFromUse(n, data, parents, ll, message);
        }else if(StringsEqual(n.type, "if".toCharArray())){
            success = GenerateDocumentFromIf(n, data, parents, ll, message);
        }else if(StringsEqual(n.type, "foreach".toCharArray())){
            success = GenerateDocumentFromForeach(n, data, parents, ll, message);
        }else if(StringsEqual(n.type, "text".toCharArray())){
            success = true;
            LinkedListCharactersAddString(ll, n.p1);
        }else if(StringsEqual(n.type, "print".toCharArray())){
            if(IsStructure(data)){
                value = GetValueFromStructOrClosestParent(data.structure, parents, n.p1, found);

                if(found.booleanValue) {
                    if(IsString(value)){
                        valueString = value.string;
                    }else {
                        valueString = WriteJSON(value);
                    }
                    LinkedListCharactersAddString(ll, valueString);
                }else{
                    success = false;
                    message.string = "Key for printing not found in JSON object: ".toCharArray();
                    message.string = ConcatenateString(message.string, n.p1);
                }
            }else{
                success = false;
                message.string = "Data structure for print command is not a JSON object.".toCharArray();
            }
        }

        return success;
    }

    public static Data GetValueFromStructOrClosestParent(Structure struct, Array parents, char [] name, BooleanReference found) {
        double i, p;
        Structure parentStruct;
        Data value;
        value = GetDataFromStructWithCheck(struct, name, found);

        if(!found.booleanValue){
            // Look in parents in reverse order.
            for(i = 0d; i < ArrayLength(parents) && !found.booleanValue; i = i + 1d){
                p = ArrayLength(parents) - 1d - i;
                parentStruct = ArrayIndex(parents, p).structure;
                value = GetDataFromStructWithCheck(parentStruct, name, found);
            }
        }
        return value;
    }

    public static boolean GenerateDocumentFromUse(Node n, Data data, Array parents, LinkedListCharacters ll, StringReference message) {
        boolean success;
        BooleanReference found;
        Data value;

        found = new BooleanReference();

        if(IsStructure(data)){
            value = GetValueFromStructOrClosestParent(data.structure, parents, n.p1, found);

            if(found.booleanValue) {
                AddStructToArray(parents, data.structure);
                success = GenerateDocumentFromNode(n.block1, value, parents, ll, message);
                ArrayRemove(parents, ArrayLength(parents));
            }else{
                success = false;
                message.string = "Key for use not found in JSON object.".toCharArray();
            }
        }else{
            success = false;
            message.string = "Data structure for use command is not a JSON object.".toCharArray();
        }

        return success;
    }

    public static boolean GenerateDocumentFromIf(Node n, Data data, Array parents, LinkedListCharacters ll, StringReference message) {
        boolean success;
        BooleanReference found;
        Data value;

        success = true;
        found = new BooleanReference();

        if(IsStructure(data)){
            value = GetValueFromStructOrClosestParent(data.structure, parents, n.p1, found);

            if(found.booleanValue){
                if(IsBoolean(value)){
                    if(value.booleanx) {
                        success = GenerateDocumentFromBlock(n.block1, data, parents, ll, message);
                    }

                    if(n.hasElseBlock){
                        if(!value.booleanx) {
                            success = GenerateDocumentFromBlock(n.block2, data, parents, ll, message);
                        }
                    }
                }else{
                    success = false;
                    message.string = "Value for if not boolean.".toCharArray();
                }
            }else{
                success = false;
                message.string = "Key for if not found in JSON object: ".toCharArray();
                message.string = ConcatenateString(message.string, n.p1);
            }
        }else{
            success = false;
            message.string = "Data structure for if command is not a JSON object.".toCharArray();
        }

        return success;
    }

    public static boolean GenerateDocumentFromForeach(Node n, Data data, Array parents, LinkedListCharacters ll, StringReference message) {
        boolean success;
        BooleanReference found;
        Data value, loopVar, arrayElement;
        double i;

        success = true;
        found = new BooleanReference();
        loopVar = CreateStructData();

        AddDataToStruct(loopVar.structure, n.p1, new Data());

        if(IsStructure(data)){
            value = GetValueFromStructOrClosestParent(data.structure, parents, n.p2, found);

            if(found.booleanValue){
                if(IsArray(value)){

                    for(i = 0d; i < value.array.length; i = i + 1d) {
                        arrayElement = ArrayIndex(value.array, i);

                        AddStructToArray(parents, data.structure);

                        if (IsStructure(arrayElement)){
                            success = GenerateDocumentFromBlock(n.block1, arrayElement, parents, ll, message);
                        }else{
                            AddDataToStruct(loopVar.structure, n.p1, arrayElement);
                            success = GenerateDocumentFromBlock(n.block1, loopVar, parents, ll, message);
                        }

                        ArrayRemove(parents, ArrayLength(parents));
                    }

                }else{
                    success = false;
                    message.string = "Value for foreach is not an array.".toCharArray();
                }

            }else{
                success = false;
                message.string = "Key for foreach not found in JSON object: ".toCharArray();
                message.string = ConcatenateString(message.string, n.p2);
            }
        }else{
            success = false;
            message.string = "Data structure for foreach command is not a JSON object.".toCharArray();
        }

        return success;
    }

    public static boolean ParseTemplate(StringReference [] tokens, Node node, StringReference message) {
        boolean success;
        NumberReference position;

        position = CreateNumberReference(0d);
        success = ParseTemplateBlock(tokens, position, node, message);

        if(success) {
            if (position.numberValue != tokens.length) {
                success = false;
                message.string = "Unexpected token at the end of template.".toCharArray();
            }
        }

        delete(position);

        return success;
    }

    public static boolean ParseTemplateBlock(StringReference [] tokens, NumberReference position, Node node, StringReference message) {
        boolean success, done;
        Node tn, nb;

        success = true;
        done = false;

        node.type = "block".toCharArray();
        node.nodes = CreateLinkedListNodes();

        for(; position.numberValue < tokens.length && success && !done;) {
            tn = new Node();
            success = ParseNodeString(tokens[(int) position.numberValue].string, tn, message);

            if(success) {
                if (StringsEqual(tn.type, "text".toCharArray()) || StringsEqual(tn.type, "print".toCharArray())) {
                    LinkedListAddNode(node.nodes, tn);
                    position.numberValue = position.numberValue + 1d;
                } else if (StringsEqual(tn.type, "use".toCharArray())) {
                    nb = new Node();
                    success = ParseUseBlock(tokens, position, nb, message);
                    LinkedListAddNode(node.nodes, nb);
                } else if (StringsEqual(tn.type, "if".toCharArray())) {
                    nb = new Node();
                    success = ParseIfBlock(tokens, position, nb, message);
                    LinkedListAddNode(node.nodes, nb);
                } else if (StringsEqual(tn.type, "foreach".toCharArray())) {
                    nb = new Node();
                    success = ParseForeachBlock(tokens, position, nb, message);
                    LinkedListAddNode(node.nodes, nb);
                } else {
                    done = true;
                }
            }
        }

        return success;
    }

    public static boolean ParseUseBlock(StringReference[] tokens, NumberReference position, Node useBlock, StringReference message) {
        boolean success;
        Node n;

        n = new Node();
        ParseNodeString(tokens[(int) position.numberValue].string, n, message);

        useBlock.type = CopyString(n.type);
        useBlock.p1 = CopyString(n.p1);
        useBlock.block1 = new Node();

        position.numberValue = position.numberValue + 1d;

        success = ParseTemplateBlock(tokens, position, useBlock.block1, message);

        if(success) {
            if(position.numberValue < tokens.length) {
                n = new Node();
                ParseNodeString(tokens[(int) position.numberValue].string, n, message);

                if (StringsEqual(n.type, "end".toCharArray())) {
                    success = true;
                    position.numberValue = position.numberValue + 1d;
                } else {
                    success = false;
                    message.string = "End block expected at the end of use block.".toCharArray();
                }
            }else{
                success = false;
                message.string = "End block expected at the end of use block.".toCharArray();
            }
        }

        return success;
    }

    public static boolean ParseIfBlock(StringReference[] tokens, NumberReference position, Node ifBlock, StringReference message) {
        boolean success;
        Node n;

        n = new Node();
        ParseNodeString(tokens[(int) position.numberValue].string, n, message);

        ifBlock.type = CopyString(n.type);
        ifBlock.p1 = CopyString(n.p1);
        ifBlock.block1 = new Node();
        ifBlock.hasElseBlock = false;

        position.numberValue = position.numberValue + 1d;

        success = ParseTemplateBlock(tokens, position, ifBlock.block1, message);

        if(success) {
            if(position.numberValue < tokens.length){
                n = new Node();
                ParseNodeString(tokens[(int) position.numberValue].string, n, message);

                if(StringsEqual(n.type, "end".toCharArray())) {
                    success = true;
                    position.numberValue = position.numberValue + 1d;
                }else if(StringsEqual(n.type, "else".toCharArray())) {
                    position.numberValue = position.numberValue + 1d;
                    ifBlock.hasElseBlock = true;
                    ifBlock.block2 = new Node();
                    success = ParseTemplateBlock(tokens, position, ifBlock.block2, message);

                    if(success) {
                        if(position.numberValue < tokens.length){
                            n = new Node();
                            ParseNodeString(tokens[(int) position.numberValue].string, n, message);

                            if(StringsEqual(n.type, "end".toCharArray())) {
                                success = true;
                                position.numberValue = position.numberValue + 1d;
                            }else{
                                success = false;
                                message.string = "End block expected at the end of else block.".toCharArray();
                            }
                        }else{
                            success = false;
                            message.string = "End block expected at the end of else block.".toCharArray();
                        }
                    }
                }else{
                    success = false;
                    message.string = "End or else block expected at the end of if block.".toCharArray();
                }
            }else{
                success = false;
                message.string = "End or else block expected at the end of if block.".toCharArray();
            }
        }

        return success;
    }

    public static boolean ParseForeachBlock(StringReference[] tokens, NumberReference position, Node foreachBlock, StringReference message) {
        boolean success;
        Node n;

        n = new Node();
        ParseNodeString(tokens[(int) position.numberValue].string, n, message);

        foreachBlock.type = CopyString(n.type);
        foreachBlock.p1 = CopyString(n.p1);
        foreachBlock.p2 = CopyString(n.p2);
        foreachBlock.block1 = new Node();

        position.numberValue = position.numberValue + 1d;

        success = ParseTemplateBlock(tokens, position, foreachBlock.block1, message);

        if(success) {
            if(position.numberValue < tokens.length){
                n = new Node();
                ParseNodeString(tokens[(int) position.numberValue].string, n, message);

                if(StringsEqual(n.type, "end".toCharArray())) {
                    success = true;
                    position.numberValue = position.numberValue + 1d;
                } else {
                    success = false;
                    message.string = "End block expected at the end of for each block.".toCharArray();
                }
            } else {
                success = false;
                message.string = "End block expected at the end of foreach block.".toCharArray();
            }
        }

        return success;
    }

    public static boolean ParseNodeString(char [] token, Node node, StringReference message) {
        char [] command;
        boolean success, isText;
        StringReference [] parts;

        success = true;
        isText = false;

        if(token.length <= 2d){
            isText = true;
        }else if(token[0] == '\\' && token[1] == '{'){
            isText = true;
        }else if(token[0] != '{'){
            isText = true;
        }else{
            command = Substring(token, 1d, token.length - 1d);
            parts = SplitByCharacter(command, ' ');

            if(command.length > 0d) {
                if (StringsEqual(parts[0].string, "use".toCharArray())) {
                    if(parts.length == 2d) {
                        node.type = CopyString(parts[0].string);
                        node.p1 = CopyString(parts[1].string);
                    }else{
                        success = false;
                        message.string = "The use command takes one parameter.".toCharArray();
                    }
                } else if (StringsEqual(parts[0].string, "end".toCharArray())) {
                    if(parts.length == 1d) {
                        node.type = CopyString(parts[0].string);
                    }else{
                        success = false;
                        message.string = "The end command takes no parameters.".toCharArray();
                    }
                } else if (StringsEqual(parts[0].string, "print".toCharArray())) {
                    if(parts.length == 2d) {
                        node.type = CopyString(parts[0].string);
                        node.p1 = CopyString(parts[1].string);
                    }else{
                        success = false;
                        message.string = "The print command takes one parameter.".toCharArray();
                    }
                } else if (StringsEqual(parts[0].string, "foreach".toCharArray())) {
                    if(parts.length == 4d) {
                        if(StringsEqual(parts[2].string, "in".toCharArray())) {
                            node.type = CopyString(parts[0].string);
                            node.p1 = CopyString(parts[1].string);
                            node.p2 = CopyString(parts[3].string);
                        }else{
                            success = false;
                            message.string = "The foreach command must have \"in\" after the first parameter.".toCharArray();
                        }
                    }else{
                        success = false;
                        message.string = "The foreach command takes three parameters.".toCharArray();
                    }
                } else if (StringsEqual(parts[0].string, "if".toCharArray())) {
                    if(parts.length == 2d) {
                        node.type = CopyString(parts[0].string);
                        node.p1 = CopyString(parts[1].string);
                    }else{
                        success = false;
                        message.string = "The if command takes one parameter.".toCharArray();
                    }
                } else if (StringsEqual(parts[0].string, "else".toCharArray())) {
                    if(parts.length == 1d) {
                        node.type = CopyString(parts[0].string);
                    }else{
                        success = false;
                        message.string = "The else command takes no parameters.".toCharArray();
                    }
                }else{
                    isText = true;
                }
            }else{
                isText = true;
            }
        }

        if(isText){
            node.type = "text".toCharArray();
            node.p1 = ReplaceString(token, "\\{print ".toCharArray(), "{print ".toCharArray());
            node.p1 = ReplaceString(node.p1, "\\{use ".toCharArray(), "{use ".toCharArray());
            node.p1 = ReplaceString(node.p1, "\\{if ".toCharArray(), "{if ".toCharArray());
            node.p1 = ReplaceString(node.p1, "\\{end}".toCharArray(), "{end}".toCharArray());
            node.p1 = ReplaceString(node.p1, "\\{foreach ".toCharArray(), "{foreach ".toCharArray());
            node.p1 = ReplaceString(node.p1, "\\{else}".toCharArray(), "{else}".toCharArray());
        }

        return success;
    }

    public static void delete(Object object) {
        // Java has garbage collection.
    }
}
